# -*- coding: utf-8 -*-
"""SBGrid Core module

This module contains core functions shared by all implementations of the SBgrid
client installer.
"""
import logging
import os.path
import platform
import shutil
import socket
import sqlite3
import subprocess
import sys
import tempfile
import threading
import time
import getpass
from ConfigParser import ConfigParser
from contextlib import closing
from tempfile import NamedTemporaryFile
from sbgrid import TARGET, CLI_NAME
import requests


def enum(**enums):
    return type('Enum', (), enums)


class Error(Exception):
    """Base class for exceptions in this module."""
    pass


class NotActivatedException(Error):
    pass


class DamagedInstallationException(Error):
    pass


class RSyncError(Error):
    """Something went wrong with rsync."""

    def __init__(self, returncode):
        super(RSyncError, self).__init__('RSync error: {0}'.format(returncode))
        self.returncode = returncode


class RSyncNotFound(RSyncError):
    pass


class Action(threading.Thread):
    """Represents interruptable unit of work."""

    def __init__(self, *args, **kwargs):
        super(Action, self).__init__(*args, **kwargs)
        self._stop = threading.Event()
        self.successful = False

    def stop(self):
        self._stop.set()

    def shouldStop(self):
        return self._stop.isSet()

    def sleep(self, seconds):
        while seconds:
            if self.shouldStop():
                break
            time.sleep(1)
            seconds = seconds - 1


def database_operation(function):
    """Decorator for database opetations."""
    def wrapper(obj, *args, **kwargs):
        conn = sqlite3.connect(obj._database_path)
        retval = function(obj, conn, *args, **kwargs)
        conn.commit()
        conn.close()
        return retval
    return wrapper


class Installation(object):
    site = None
    user = None
    key = None

    grids = None
    collections = None
    packages = None

    debug = True

    def __init__(self, api='sbgrid.org/api', rsync='sync2.sbgrid.org', target=TARGET, link='/programs',
                 config_path=os.path.expanduser('~/.{0}_installer'.format(CLI_NAME))):
        self.api = api
        self.rsync = rsync
        self.target = target
        self.link = link
        self._config_path = config_path
        self._raw = None

        if sys.platform == 'darwin':
            self.platform = 'i386-mac'
        elif sys.platform.startswith('linux'):
            self.platform = '{0}-linux'.format(platform.machine())
        else:
            raise Exception('Unsupported platform {0}!'.format(sys.platform))

    @property
    def _api_url(self):
        return 'https://{0}/installation/{1}/'.format(self.api, self.site)

    @property
    def _database_path(self):
        return self._share_path('packages-{0}.db'.format(self.site))

    @database_operation
    def _load_db_status(self, conn, packages, version):
        return conn.execute('''SELECT default_version, timestamp, status FROM packages
                               WHERE slug=? AND version=?''', (packages.slug, version.version)).fetchone()

    def _load(self):
        """Fetch and merge data from API and database into data model objects.
        """
        if self._raw is None:
            logging.debug('Calling service: GET %s', self._api_url)
            response = requests.get(self._api_url)
            response.raise_for_status()
            logging.debug('Elapsed: %s seconds', response.elapsed.total_seconds())
            self._raw = response.json()[0]
        self.grids = list()
        self.collections = dict()
        for grid in self._raw['grid']:
            self.grids.append(grid['name'])
            self.collections[grid['name']] = [c['collection'] for c in
                                              sorted(grid['collection'], key=lambda c: c['collection_grid__position'])]
        # Repackage information from API into model objects
        if not self.packages:
            self.packages = list()
            for key in ('software_available', 'software_inclusions'):
                if key in self._raw:
                    for package in self._raw[key]:
                        # Ignore packages with no versions for current platform
                        if self.platform in {v['os_dir_name'] for v in package['available_os']}:
                            # TODO this dirty hack should be removed after switching to API v2!
                            package['grid'] = filter(lambda g: g['name'] in self.grids, package['grid'])
                            self.packages.append(Package(package, self.platform))
        # Load status from DB
        for package in self.packages:
            for version in package.versions:
                result = self._load_db_status(package, version)
                if result is None:
                    version.installed_default, version.installed_timestamp, version.status = (False, None, None)
                else:
                    version.installed_default, version.installed_timestamp, version.status = result
                    version.installed_default = bool(version.installed_default)
        # Calculate updates
        for package in filter(lambda p: p.installed, self.packages):
            for version in filter(lambda v: v.installed, package.versions):
                logging.debug('Checking %s %s', package.slug, version.version)
                if (version.installed_default and not version.default and
                        version.version != package.default_version.version):
                    logging.debug('Found upgrade to %s', package.default_version.version)
                    version.update = True
                    version.new_version = package.default_version
                    continue
                if version.timestamp > version.installed_timestamp:
                    logging.debug('Found upgrade from %s to %s', version.installed_timestamp, version.timestamp)
                    version.update = True
                    version.new_version = None

    def authenticate(self):
        logging.debug('Authenticating user %s for site %s', self.user, self.site)
        url = 'rsync://{0}/{1}/'.format(self._rsync_server, self.site)
        command = [self._rsync_path, '--list-only']
        if logging.getLogger().getEffectiveLevel() == logging.DEBUG:
            command += ['--verbose']
        else:
            command += ['--quiet']
        command += [url]
        logging.debug('Running %s', ' '.join(command))
        try:
            _rsync_exec(command, dict(USER=self.user, RSYNC_PASSWORD=self.key))
            return True
        except:
            logging.debug('Authentication to RSync server failed!', exc_info=self.debug)
            return False

    def activate(self, site, user, key):
        """Activate installation.

        This function bootstraps (when necessary) and configures installation using provided account.

        Args:
            site (str): Site name.
            user (str): User name.
            key (str): Activation key.

        Return:
            bool: True if provided with correct site name and valid activation key,
                  False otherwise.

        """
        self.site = site
        self.user = user
        self.key = key

        if not self.authenticate():
            logging.warning('Key validation failed for user: %s', self.user)
            return False

        self._create_directories()
        self._synchronize([self.platform,
                           os.path.join(self.platform, 'system'),
                           os.path.join(self.platform, 'system', 'rsync'),
                           os.path.join(self.platform, 'system', 'rsync', '**')])
        self._synchronize(['/share',
                           '/share/sbgrid',
                           '/share/sbgrid/**',
                           '/share/grids',
                           '/share/grids/**'])
        self._initialize_package_database()

        # Set site name
        import sbgrid
        sitename = '{0}_{1}'.format(self.site, sbgrid.__version__)
        sitename_path = self._share_path('sitename')
        if (not os.path.exists(sitename_path) or open(sitename_path, 'r').readline().strip() != sitename):
            with open(sitename_path, 'w') as f:
                f.write(sitename)
                f.write('\n')

        try:
            self._load()
        except:
            logging.error('Failed to load installation data for site: %s', self.site, exc_info=self.debug)
            return False
        return True

    def load(self):
        """Load config file and packages database."""
        config = ConfigParser()
        if not os.path.isfile(self._config_path):
            raise NotActivatedException('Installation config does not exist. Please activate your installation.')
        config.read(self._config_path)

        if not config.has_section('installer'):
            raise NotActivatedException('Invalid config. Please reactivate your installation.')
        for key, value in config.items('installer'):
            setattr(self, key, value)

        # No target dir -> not activated
        if not os.path.exists(self.target):
            raise NotActivatedException('Main installation directory is missing! Please reactivate your installation.')

        if not os.path.exists(self._platform_path('system', 'rsync', 'current', 'bin', 'rsync')):
            self._synchronize([self.platform,
                               os.path.join(self.platform, 'system'),
                               os.path.join(self.platform, 'system', 'rsync'),
                               os.path.join(self.platform, 'system', 'rsync', '**')])

        # Missing vital parts -> damaged
        if not os.path.islink(self.link):
            raise DamagedInstallationException('Link to installation folder is broken!')

        if not os.path.exists(self._database_path):
            self._initialize_package_database()

        self._load()

        # Pending removals -> try to fix
        need_update = False
        for package in self.packages:
            for version in package.versions:
                if version.status == Package.STATUS.REMOVED:
                    need_update = True
                    logging.debug('Retrying removal %s version %s', package.slug, version.version)
                    self.remove(package, version)
        if need_update:
            self.setup()
            self._load()

        # Pending installations -> user decision required
        if self.pending_packages():
            logging.warn('Found pending installations:')
            for slug, version in self.pending_packages():
                logging.warn('%s %s', slug, version)
            logging.warn('Try `resume` command.')

        # Missing packages -> reinstall?
        if self.missing_packages():
            logging.warn('There are some missing packages')

    def save(self):
        """Save current configuration to config file."""
        config = ConfigParser()
        config.add_section('installer')
        for key in ['site', 'user', 'key']:
            config.set('installer', key, getattr(self, key))
        with open(self._config_path, 'w') as f:
            config.write(f)

    @staticmethod
    def check_if_expired(site):
        """Check installation status and return expirity message if available.

        Return:
            str: Expirity message or None.
        """
        try:
            logging.debug('Checking if site %s expired...', site)
            response = requests.get('http://sync.sbgrid.org/expire/{0}'.format(site))
            logging.debug('Elapsed: %s seconds', response.elapsed.total_seconds())
            if response.status_code == 200:
                logging.critical('Site EXPIRED: %s', response.text)
                return response.text
            logging.debug('Site OK')
        except:
            logging.warn('Failed to check if installation expired!')
        return None

    def _create_directories(self):
        if os.path.exists(self.target) or os.path.exists(self.link):
            raise Exception('Unable to continue! Please remove: \'{0}\' and \'{1}\''.format(self.target, self.link))
        logging.debug('Creating directories for target %s and link %s', self.target, self.link)
        create_basedir_script = NamedTemporaryFile(delete=False)
        create_basedir_script.write("""if [[ ! -d {target} ]]; then
        mkdir -p {target}
    fi
    chown -R {uid}:{gid} {target}
    if [[ ! -L {link} ]]; then
        ln -s {target} {link}
    fi
    chown -h {uid}:{gid} {link}
    """.format(target=self.target, link=self.link, uid=os.geteuid(), gid=os.getegid()))
        create_basedir_script.flush()
        create_basedir_script.close()

        script_command = [_which('bash'), create_basedir_script.name]
        sudo_command = None
        if sys.stdout.isatty():
            logging.debug('Running sudo from tty')
            if not _which('sudo'):
                raise Exception('"sudo" is required to bootstrap installation!')
            sudo_command = [_which('sudo'), '-p',
                            '[sudo] password for {0}: '.format(getpass.getuser())] + script_command
        else:
            logging.debug('Attempting graphical sudo')
            if sys.platform.startswith('linux'):
                if _which('pkexec'):
                    sudo = _which('pkexec')
                elif _which('gksudo'):
                    sudo = _which('gksudo')
                elif _which('kdesudo'):
                    sudo = _which('kdesudo')
                else:
                    logging.fatal('No graphical "sudo" equivalent found!')
                    raise Exception('No graphical "sudo" equivalent found!')
                logging.debug('Using sudo helper: %s', sudo)
                sudo_command = [sudo] + script_command
            elif sys.platform == 'darwin':
                sudo_command = ['osascript', '-e', 'do shell script "bash {0}" with administrator privileges'.format(create_basedir_script.name)]  # noqa
            else:
                raise Exception('Unsupported platform {0}!'.format(sys.platform))
        if sudo_command:
            subprocess.call(sudo_command)
        os.unlink(create_basedir_script.name)

        for path in [self.platform, 'share']:
            path = os.path.join(self.target, path)
            if not os.path.exists(path):
                logging.debug('Creating directory: %s', path)
                os.mkdir(path)

    @database_operation
    def _initialize_package_database(self, conn):
        logging.debug('Initializing package database')
        conn.execute('''CREATE TABLE packages (
            slug TEXT,
            version TEXT,
            default_version BOOL,
            timestamp INTEGER,
            install_timestamp INTEGER,
            status TEXT
        )''')

    def _platform_path(self, *args):
        if args:
            return os.path.join(self.target, self.platform, *args)
        return os.path.join(self.target, self.platform)

    @property
    def _rsync_server(self):
        """Return rsync server and port."""
        if _is_port_open(self.rsync, 873):
            return '{0}:{1}'.format(self.rsync, 873)
        elif _is_port_open(self.rsync, 8080):
            return '{0}:{1}'.format(self.rsync, 8080)
        raise RSyncError('Unable to conncet to rsync server!')

    @property
    def _rsync_path(self):
        rsync_path = self._platform_path('system', 'rsync', 'current', 'bin', 'rsync')
        if os.path.isfile(rsync_path):
            return rsync_path
        return _which('rsync')

    def setup(self):
        """Update base installation and run createrc.

        Returns:
            bool: True if successful (createrc returned 20), False otherwise.

        """
        def run(path, opts, env=None, ok_code=0):
            path = os.path.join(self.target, *path)
            command = [path, opts, self.target]
            logging.debug('Running %s', ' '.join(command))
            p = subprocess.Popen(command, env=env, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            stdout, stderr = p.communicate()
            logging.debug(stdout)
            if p.returncode != ok_code:
                logging.debug(stderr)
                raise Error('Command %s reutrned %d'.format(
                    command, p.returncode))

        logging.info('Updating installation configuration...')
        try:
            self._synchronize_includes(self._share_path('sbgrid', 'lib', '{0}.includes'.format(self.platform)))
            env = None
            if logging.getLogger().getEffectiveLevel() == logging.DEBUG:
                env = dict(script_debug='1')
            run(['share', 'sbgrid', 'bin', 'createrc'], '-ft', env, 20)
            if logging.getLogger().getEffectiveLevel() == logging.DEBUG:
                env = dict(ASSEMBLE_LOG='/dev/stdout')
            run(['share', 'capsules', 'sbin', 'assemble_local'], '-t', env)
            shrc = self._share_path('grids', '.admin', 'allgrid.shrc')
            if os.path.exists(shrc):
                logging.debug('Sourcing %s', shrc)
                try:
                    subprocess.Popen(['/bin/bash', shrc], stdout=open(os.devnull, 'wb'))
                except:
                    logging.debug('There where errors sourcing %s!', shrc, exc_info=self.debug)
        except RSyncError, e:
            logging.debug('RSync failed when updating base installation! Error code: %s', e.returncode)
        except Exception, e:
            logging.error(e.message)
            return False
        return True

    def _rsync_w_retry(self, opts, path, dest):
        url = 'rsync://{0}/{1}/{2}'.format(self._rsync_server, self.site, path)
        rsync_command = [self._rsync_path] + opts + [url, dest]
        logging.debug('Running %s', ' '.join(rsync_command))

        return_code = None
        for backoff in (0, 20, 40, 60, 60):
            if backoff:
                logging.debug('Backoff: %s', backoff)
                if hasattr(threading.current_thread(), 'sleep'):
                    threading.current_thread().sleep(backoff)
                else:
                    time.sleep(backoff)
            if hasattr(threading.current_thread(), 'shouldStop') and threading.current_thread().shouldStop():
                break
            return_code = _rsync_exec(rsync_command, dict(USER=self.user, RSYNC_PASSWORD=self.key))
            # if interrupted or successful stop retrying
            if return_code is None or return_code == 0:
                break
        return return_code

    def _synchronize(self, includes):
        with tempfile.NamedTemporaryFile() as f:
            for line in includes:
                f.write(line)
                f.write('\n')
            f.flush()
            self._synchronize_includes(f.name)

    def _synchronize_includes(self, includes_file):
        opts = ['--compress',
                '--delete-after',
                '--devices',
                '--links',
                '--omit-dir-times',
                '--perms',
                '--specials',
                '--times',
                '--recursive',
                '--include-from={0}'.format(includes_file),
                '--exclude=*']
        if logging.getLogger().getEffectiveLevel() == logging.DEBUG:
            opts.append('--verbose')
        self._rsync_w_retry(opts, 'main/programs/', self.target)

    def package(self, slug):
        for package in self.packages:
            if package.slug == slug:
                return package
        raise Exception('Package not found!')

    def get_version(self, slug, version):
        package = self.package(slug)
        for package_version in package.versions:
            if package_version.version == version:
                return package_version
        raise Exception('Version {0} for package {1} not found'.format(version, slug))

    def packages_by_collection(self, grid, collection):
        if grid not in self.grids:
            raise Exception('Unknown grid name!')
        if collection not in self.collections[grid]:
            raise Exception('Unknown collection name!')
        return filter(lambda p: collection in p.collections and grid in p.grids, self.packages)

    def pending_packages(self):
        pending = list()
        for package in self.packages:
            for version in package.versions:
                if version.status == Package.STATUS.PENDING:
                    pending.append((package.slug, version.version))
        return pending

    def missing_packages(self):
        missing = list()
        for package in self.packages:
            for version in package.versions:
                if version.installed:
                    if not os.path.exists(self._platform_path(package.installation_directory, version.version)):
                        missing.append((package.slug, version.version))
        return missing

    def installed_packages(self):
        return filter(lambda p: p.installed, self.packages)

    def get_packages_to_remove(self):
        def can_remove_the_package(package, version):
            if not version.installed:
                return False
            # checking if version is default
            if version.default:
                return False
            # checking if the highest version
            if not package.default_version.installed:
                if package.installed_version[-1] == version:
                    return False
            return True

        packages_to_remove = list()
        versions_to_remove = list()
        installation_to_save = list()
        # checking witch package should stay
        for package in self.installed_packages():
            for version in package.versions:
                if version.installed:
                    if not can_remove_the_package(package, version):
                        self._get_necessary_package_depenencies(version, installation_to_save)
                        installation_to_save.append('{0} {1}'.format(package.slug, version))
        # creating list o packages to remove
        for package in self.installed_packages():
            for version in package.versions:
                if version.installed or version.pending:
                    if '{0} {1}'.format(package.slug, version) not in installation_to_save:
                        packages_to_remove.append(package)
                        versions_to_remove.append(version)
        return zip(packages_to_remove, versions_to_remove)

    def _get_necessary_package_depenencies(self, version, necessary_dependencies):
        if not version.depends_on:
            return necessary_dependencies
        for dependency in version.depends_on:
            necessary_dependencies.append('{0} {1}'.format(dependency[0], dependency[1]))
            version = self.get_version(dependency[0], dependency[1])
            if version.depends_on:  # in case if dependency has own dependency
                self._get_necessary_package_depenencies(version, necessary_dependencies)
        return version

    def install_by_name(self, slug, ver):
        """Install specific version of the package by name.

        Args:
            slug (str): Pacakge name slug.
            ver (str): Version string.
        """
        package = self.package(slug)
        self.install(package, package.version(ver))

    def install(self, package, version):
        self._set_status(package, version, Package.STATUS.PENDING)
        # check if package is installed, if there are previous versions available reuse them
        if not version.installed:
            installed_by_timestamp = sorted(filter(lambda v: v.installed, package.versions),
                                            key=lambda v: v.timestamp, reverse=True)
            if installed_by_timestamp:
                most_recent = installed_by_timestamp[0]
                if not os.path.islink(self._platform_path(package.installation_directory, most_recent.version)):
                    logging.debug('Found version %s of the package %s', most_recent.version, package.slug)
                    self._copytree(self._platform_path(package.installation_directory, most_recent.version),
                                   self._platform_path(package.installation_directory, version.version))

                # TODO Reuse share directory?!?
        try:
            # rsync package
            self._synchronize(self._includes_for(package, version))
            self._set_status(package, version, Package.STATUS.INSTALLED)
        except RSyncError, e:
            logging.error(e.message, exc_info=self.debug)
            self.remove(package, version)
        self._load()

    def install_with_dependencies_from_name(self, slug, version):
        pkg = self.package(slug)
        version = pkg.version(version)
        self.install_with_dependencies(pkg, version)

    def install_with_dependencies(self, package, version):
        for s, v in version.depends_on:
            pkg = self.package(s)
            self.install_with_dependencies(pkg, pkg.version(v))
        self.install(package, version)

    @database_operation
    def _set_status(self, conn, package, version, status):
        # Update model
        version.status = status
        # Update database
        curr = conn.execute('''SELECT slug FROM packages WHERE slug=? AND version=?''', (package.slug, version.version))
        if curr.fetchone() is None:
            conn.execute('''INSERT INTO packages VALUES
                         (?, ?, ?, ?, strftime('%s','now'), ?)''',
                         (package.slug, version.version, version.default, version.timestamp, status))
        else:
            conn.execute('''UPDATE packages
                         SET default_version=?, timestamp=?, install_timestamp=strftime('%s','now'), status=?
                         WHERE slug=? AND version=?''',
                         (version.default, version.timestamp, status, package.slug, version.version))
        logging.debug('Status for %s %s (default: %s, timestamp: %s) set to %s',
                      package.slug, version.version, version.default, version.timestamp, status)

    def _includes_for(self, package, version):
        """Return list of rsync includes and directories that have to be synchronized to install package."""
        # List package root using rsync
        package_dir_contents = self._rsync_list(os.path.join(self.platform, package.installation_directory))
        # Filter files - whitelist vs all file?
        includes = map(lambda i: i['path'], filter(lambda i: i['type'] == 'file', package_dir_contents.values()))
        # Add dirs to enable traversal for rsync
        includes.append(os.path.join('/', self.platform))
        includes.append(os.path.join('/', self.platform, package.installation_directory))
        # Add capsules if relevant
        if '.manifest' in package_dir_contents:
            includes.append(package_dir_contents['.manifest']['path'])
            includes.append(os.path.join(package_dir_contents['.manifest']['path'], '**'))
        includes.append(package_dir_contents[version.version]['path'])
        # Is version a link of dir? Ignore if it is a link.
        if package_dir_contents[version.version]['type'] == 'directory':
            includes.append(os.path.join(package_dir_contents[version.version]['path'], '**'))
        if package.has_share_directory:
            # List package share dir
            share_dir = os.path.join('share', package.installation_directory)
            includes.append('/share')
            includes.append(os.path.join('/', share_dir))
            share_dir_contents = self._rsync_list(share_dir)
            if version.version in share_dir_contents:
                # Use versioned share dir
                includes.append(share_dir_contents[version.version]['path'])
                includes.append(os.path.join(share_dir_contents[version.version]['path'], '**'))
            else:
                # Use unversioned share dir
                includes.append(os.path.join('/', share_dir, '**'))
        return includes

    def remove_package(self, package):
        for version in package.versions:
            if version.installed:
                self._set_status(package, version, Package.STATUS.REMOVED)
        self._rm_tree(self._platform_path(package.installation_directory))
        if package.has_share_directory:
            self._rm_tree(self._share_path(package.installation_directory))
        self._remove_status_for_package(package)

    def remove(self, package, version):
        if not (version.installed or version.pending):
            raise Exception('Package {0} version {1} is not installed!'.format(package.slug, version.version))
        self._set_status(package, version, Package.STATUS.REMOVED)
        if not package.installed:
            logging.debug('Removing last version of %s', package.slug)
            self.remove_package(package)
            return

        logging.debug('Removing package %s version %s', package.slug, version.version)
        # is install directory a link?
        version_path = self._platform_path(package.installation_directory, version.version)
        if not os.path.islink(version_path):
            self._rm_tree(version_path)
        else:
            target_path = os.readlink(version_path)
            os.unlink(version_path)
            if self._can_remove_target(self._platform_path(package.installation_directory), target_path):
                self._rm_tree(target_path)
        if package.has_share_directory:
            # is share directory a link?
            versioned_share_path = self._share_path(package.installation_directory, version.version)
            if not os.path.islink(versioned_share_path):
                self._rm_tree(versioned_share_path)
            else:
                target_path = os.readlink(versioned_share_path)
                os.unlink(versioned_share_path)
                if self._can_remove_target(self._share_path(package.installation_directory), target_path):
                    self._rm_tree(target_path)
        self._remove_status_for_version(package, version)

    def _share_path(self, *args):
        if args:
            return os.path.join(self.target, 'share', *args)
        return os.path.join(self.target, 'share')

    @database_operation
    def _remove_status_for_package(self, conn, package):
        conn.execute('DELETE FROM packages WHERE slug=?', (package.slug,))

    @database_operation
    def _remove_status_for_version(self, conn, package, version):
        conn.execute('DELETE FROM packages WHERE slug=? AND version=?', (package.slug, version.version))

    def _rm_tree(self, path):
        try:
            logging.debug('Removing %s', path)
            shutil.rmtree(path)
        except OSError:
            logging.debug('Failed to remove dir: %s', path, exc_info=self.debug)

    def _can_remove_target(self, basedir, target):
        for name in os.listdir(basedir):
            path = os.path.join(basedir, name)
            if os.path.islink(path):
                if os.path.join(basedir, os.readlink(path)) == target:
                    return False
        return True

    def _rsync_list(self, path):
        url = 'rsync://{0}/{1}/{2}'.format(self._rsync_server, self.site, os.path.join('main', 'programs', path, ''))
        rsync_command = [self._rsync_path, '--list-only', '--links', url]
        logging.debug('Running %s', ' '.join(rsync_command))
        env = dict(USER=self.user, RSYNC_PASSWORD=self.key)
        p = subprocess.Popen(rsync_command, env=env, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = p.communicate()
        if p.returncode != 0:
            logging.debug('RSync returned: %d', p.returncode)
            logging.debug(stderr)
            return None

        items = dict()
        for line in stdout.split('\n'):
            columns = line.split()
            if not columns or columns[4] == '.':
                continue
            item = dict()
            if line[0] == 'l' and columns[5] == '->':
                # eg.: lrwxrwxrwx             32 2016/09/02 19:29:53 0.16.2 -> /programs/i386-mac/python/2.7.10
                item['type'] = 'link'
                # is this relative or absolute link?
                if columns[6].startswith('/programs'):
                    # NOTE those links are handled by dependencies - do not follow!
                    item['destination'] = os.path.join('/', path, columns[6][9:])
                else:
                    item['destination'] = os.path.join('/', path, columns[6])
            else:
                if line[0] == 'd':
                    # eg.: drwxr-xr-x          17 2011/08/02 21:14:21 11.02
                    item['type'] = 'directory'
                else:
                    # eg.: -rw-r--r--         167 2014/03/13 23:16:37 LICENSE
                    item['type'] = 'file'
            item['path'] = os.path.join('/', path, columns[4])
            items[columns[4]] = item
        return items

    def _copytree(self, src, dst):
        """Recursively copy directory using tree.
        """
        rsync_command = [self._rsync_path, '--archive', os.path.join(src, ''), dst]
        logging.debug('Running %s', ' '.join(rsync_command))
        p = subprocess.Popen(rsync_command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = p.communicate()
        if p.returncode != 0:
            logging.debug('RSync returned: %d', p.returncode)
            logging.debug(stderr)

    def rebuild(self):
        logging.debug('Attempting full installation rebuild, this might take a while...')
        includes = set()
        for package in self.installed_packages():
            for version in package.versions:
                if version.installed:
                    includes.update(self._includes_for(package, version))
        logging.debug('Includes to be synchronized:')
        for include in includes:
            logging.debug(include)
        try:
            self._synchronize(includes)
            return True
        except RSyncError, e:
            logging.error(e.message, exc_info=self.debug)
        return False

    def updated_packages(self):
        return filter(lambda p: p.update, self.packages)

    def update(self, package, version):
        if version.new_version is None:
            logging.info('In place update of %s %s from release %d to %d', package.slug, version.version,
                         version.installed_timestamp, version.timestamp)
            self.install(package, version)
        else:
            logging.info('Update of %s from version %s to %s', package.slug, version.version,
                         version.new_version.version)
            self.install(package, version.new_version)
            self.remove(package, version)
        version.update = False
        version.new_version = None


class Package(object):

    STATUS = enum(PENDING='pending',
                  INSTALLED='installed',
                  REMOVED='removed')

    def __init__(self, raw, platform):
        self.slug = raw['slug']
        self.title = raw['title']
        self.active = raw['active']
        self.description = raw.get('description', '')
        self.links = list()
        for key in ['website', 'manual', 'forum_help']:
            if key in raw and raw[key]:
                self.links.append((key.replace('_', ' ').capitalize(), raw[key]))
        self.technical_notes = raw.get('technical_notes', '').strip()
        if self.technical_notes:
            self.technical_notes = self.technical_notes.replace('\r', ' ').replace('\n', ' ')
        self.installation_directory = raw['installation_directory'].strip()
        if not self.installation_directory:
            self.installation_directory = self.slug
        self.has_share_directory = raw['share_dir']
        self.collections = raw['collection']
        self.grids = [g['name'] for g in raw['grid']]

        self.versions = list()
        for version in raw.get('available_os', []):
            if version['os_dir_name'] == platform:
                self.versions.append(Version(version))

        # Ugly hack for packages with no default version?
        if self.default_version is None:
            self.versions[0].default = True

    def __str__(self):
        return self.slug

    def __repr__(self):
        return str(self)

    @property
    def installed(self):
        return any([v.installed for v in self.versions])

    @property
    def update(self):
        return any([v.update for v in self.versions])

    @property
    def updated_versions(self):
        return filter(lambda v: v.update, self.versions)

    @property
    def default_version(self):
        default = filter(lambda v: v.default, self.versions)
        if default:
            return default.pop()
        return None

    @property
    def installed_version(self):
        return sorted(filter(lambda v: v.installed, self.versions), key=lambda v: v.timestamp, reverse=False)

    def version(self, version):
        for v in self.versions:
            if v.version == version:
                return v


class Version(object):
    installed_timestamp = None
    installed_default = False
    status = None
    update = False
    new_version = None

    def __init__(self, raw):
        self.version = raw['version']
        self.default = raw['default']
        self.obsolete = raw['obsolete']
        self.timestamp = int(raw['timestamp'])
        self.depends_on = list()
        for dependency in raw.get('depends_on', []):
            self.depends_on.append((dependency['software'], dependency['version']))

    @property
    def installed(self):
        return self.status == Package.STATUS.INSTALLED

    @property
    def pending(self):
        return self.status == Package.STATUS.PENDING

    def __str__(self):
        return self.version


def _which(program):
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file
    raise Exception('{} command not found!'.format(program))


def _is_port_open(host, port):
    """Test if port is opened or closed/blocked."""
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
        sock.settimeout(2)
        if sock.connect_ex((host, port)) == 0:
            return True
        return False


def _rsync_exec(cmd, env):
    process = subprocess.Popen(cmd, env=env, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    while True:
        logging.debug(process.stdout.readline().strip())
        if process.poll() is not None:
            logging.debug('RSync exited with returncode: %s', process.returncode)
            if process.returncode != 0:
                if process.returncode == 23:
                    raise RSyncNotFound(process.returncode)
                raise RSyncError(process.returncode)
            return
        if hasattr(threading.current_thread(), 'shouldStop') and threading.current_thread().shouldStop():
            logging.debug('Terminating subprocess')
            process.terminate()
            process.wait()
            logging.debug('RSync terminated!')
            raise RSyncError(-1)
